;;; $Id: translate-module-syntax.scm,v 1.7 2006/01/18 23:17:27 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Translate MzScheme modules into Chez Scheme
;;; AUTHOR: Gary T. Leavens, with critical help from R. Kent Dyvbig

;; Set the following to #f to not load only when necessary.
;; The default is true, so that each module only has one set
;; of locations for variables local to that module.
(define *trms:load-only-if-needed* #t)

;; The variable *trms:collects-dir* comes from the file chez-localize.scm.
;; You need to edit that file to define it.
;; It has to be loaded before this file.

(define trms:load-if-needed
  (lambda (fpath)
    (let ((careful-load
	   (lambda (f)
	     (if (file-exists? fpath)
		 (load fpath)
		 (begin
		   (error 'require
			  (string-append "required file `"
					 f "' does not exist")))))))
      (if *trms:load-only-if-needed*
	  (let ((mod-sym (trms:extract-name fpath)))
	    (unless (getprop mod-sym 'require)
		    (careful-load fpath)
		    (putprop mod-sym 'require #t)))
	  (careful-load fpath)))))


;; (deftype trms:extract-name (-> (string) symbol)
(define trms:extract-name
  (letrec ((trim-extension
            (lambda (str)
              (let ((dot-pos (last-pos str #\. (- (string-length str) 1))))
                (if (= dot-pos -1)
                    str
                    (substring str 0 dot-pos)))))
           (trim-directory
            (lambda (str)
              (let* ((len (string-length str))
                     (slash-pos (last-pos str #\/ (- len 1))))
                (if (= slash-pos -1)
                    str
                    (substring str (+ slash-pos 1) len)))))
           (last-pos
            (lambda (str ch pos)
              (if (< pos 0)
                  -1
                  (if (char=? ch (string-ref str pos))
                      pos
                      (last-pos str ch (- pos 1)))))))
    (lambda (str)
      ;; ENSURES: result is a module name taken from the filename part
      ;; of the given string without the suffix.
      (string->symbol (trim-extension (trim-directory str))))))

;; (deftype trms:add-slashes (-> ((list-of string)) string))
(define trms:add-slashes
  (letrec
      ((iter
        (lambda (los ans)
          (if (null? los)
              ans
              (iter (cdr los) (string-append ans (car los) "/"))))))
    (lambda (los)
      ;; ENSURES: result is a string with the strings in los, in
      ;; order, terminated by a slash "/" character.
      (iter los ""))))

;; (deftype trms:all (forall (T) (-> ((-> (T) boolean) (list-of T)) boolean)))
(define trms:all
  (lambda (pred ls)
    ;; ENSURES: result is true just when pred is true for every element of ls
    (or (null? ls)
        (and (pred (car ls)) (trms:all pred (cdr ls))))))

;; For the module syntax,
;; we use trms:in-process (see below) to handle require, and defrep.
;; Without this, it's not possible to deal with require properly,
;; as just using the translation of require itself makes it so
;; Chez Scheme doesn't see the definitions being exported.

(define-syntax module
  (let ()
    ;; The idea of importing scheme here is from Kent Dyvbig.  This
    ;; makes it so inside this scope, the translation "module" refers to
    ;; the Chez Scheme module, not to the syntax being translated.
    (import scheme)
    (syntax-rules (provide require defrep deftype require-for-syntax)
      ((module module-identifier initial-required-module-name
	       (provide exported-id ...)
	       (require-for-syntax require-spec ...)
	       form ...)
       (trms:in-process module-identifier initial-required-module-name
			(provide exported-id ...)
			(require-for-syntax require-spec ...)
			form ...))
      ((module module-identifier initial-required-module-name
	       (provide exported-id ...)
	       (deftype name type)
	       form ...)
       (trms:in-process module-identifier initial-required-module-name
			(provide exported-id ...)
			form ...))
      ((module module-identifier initial-required-module-name
	       (provide exported-id ...)
	       (defrep abs-conc-pair ...)
	       (require require-spec ...)
	       form ...)
       (error 'module "The defrep must come after the require in a module"))
      ((module module-identifier initial-required-module-name
	       (provide exported-id ...)
	       (require require-spec ...)
	       (defrep abs-conc-pair ...)
	       form ...)
       (trms:in-process module-identifier initial-required-module-name
			(provide exported-id ...)
			(require require-spec ...)
			form ...))
      ((module module-identifier initial-required-module-name
	       (provide exported-id ...)
	       (require require-spec ...)
	       form ...)
       (trms:in-process module-identifier initial-required-module-name
			(provide exported-id ...)
			(require require-spec ...)
			form ...))
      ((module module-identifier initial-required-module-name
	       (provide exported-id ...)
	       (defrep abs-conc-pair ...)
	       form ...)
       (module module-identifier (exported-id ...)
	       form ...))
      ((module module-identifier initial-required-module-name
	       (provide exported-id ...)
	       form ...)
       (module module-identifier (exported-id ...)
	       form ...)))))

(define-syntax trms:in-process
  (lambda (x)
   (import scheme) ; see the comment in module above
   (syntax-case x  (provide require deftype defrep require-for-syntax)
     ;; take out deftypes after provide (may be before require or defrep)
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (deftype name type)
		       form ...)
      (syntax (trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       form ...)))
     ;; take out require-for-syntax
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require-for-syntax require-spec ...)
		       (require-for-syntax require-spec2 ...)
		       form ...)
      (error 'module "There can be at most one require-for-syntax form in a module"))
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require-for-syntax require-spec ...)
		       form ...)
      (syntax (trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       form ...)))
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (defrep abs-conc-pair ...)
		       (require require-spec ...)
		       form ...)
       (error 'module "The defrep must come after the require in a module"))
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require require-spec ...)
		       (require require-spec2 ...)
		       form ...)
       (error 'module "There can be at most one require form in a module"))
     ;; take out defreps following require
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require require-spec ...)
		       (defrep abs-conc-pair ...)
		       form ...)
      (syntax (trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require require-spec ...)
		       form ...)))
     ;; take out defrep (if no require)
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (defrep abs-conc-pair ...)
		       form ...)
      (syntax (trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       form ...)))
     ;; done if nothing in require
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require )
		       form ...)
      (syntax (module module-identifier (exported-id ...)
                      form ...)))
     ;; require-spec of the form (file path)
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require (kwd path) require-spec ...)
		       form ...)
      (and (literal-identifier=? (syntax kwd) (syntax file))
	   (string? (syntax-object->datum (syntax path))))
      (let ((mod-sym
             (trms:extract-name
              (syntax-object->datum (syntax path))))
            (path-str (syntax-object->datum (syntax path))))
        (with-syntax
         ((modname (datum->syntax-object (syntax kwd) mod-sym)))
         ;; have to load at expansion time
         (trms:load-if-needed path-str)
         (syntax (trms:in-process module-identifier
			     initial-required-module-name	  
			     (provide exported-id ...)
                             (require require-spec ...)
                             (import modname)
                             form ...)))))
     ;; pretty-print is already built in to Chez Scheme (hack)
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require (kwd fname collect) require-spec ...)
		       form ...)
      (and (literal-identifier=? (syntax kwd) (syntax lib))
	   (string=? "pretty.ss" (syntax-object->datum (syntax fname)))
	   (string=? "mzlib" (syntax-object->datum (syntax collect))))
      (syntax (trms:in-process module-identifier
			       initial-required-module-name
			       (provide exported-id ...)
			       (require require-spec ...)
			       form ...)))
     ;; require-spec of the form (lib fname collect ...)
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require (kwd fname collect ...) require-spec ...)
		       form ...)
      (and (literal-identifier=? (syntax kwd) (syntax lib))
	   (string? (syntax-object->datum (syntax fname)))
           (trms:all string?
                     (map syntax-object->datum (syntax (collect ...)))))
      (let ((mod-sym
             (trms:extract-name
              (syntax-object->datum (syntax fname))))
            (path-str
             (string-append
	      *trms:collects-dir* "/"
              (trms:add-slashes
               (syntax-object->datum (syntax (collect ...))))
              (syntax-object->datum (syntax fname)))))
        (with-syntax
         ((modname (datum->syntax-object (syntax kwd) mod-sym)))
         ;; have to load at expansion time
         (trms:load-if-needed path-str)
         (syntax (trms:in-process module-identifier
				  initial-required-module-name
				  (provide exported-id ...)
				  (require require-spec ...)
				  (import modname)
				  form ...)))))
     ;; require-spec of the form "path"
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require path require-spec ...)
		       form ...)
      (string? (syntax-object->datum (syntax path)))
      (let ((mod-sym
             (trms:extract-name
              (syntax-object->datum (syntax path))))
            (path-str (syntax-object->datum (syntax path))))
        (with-syntax
         ((modname
           (datum->syntax-object (syntax require) mod-sym)))
         ;; have to load at expansion time
         (trms:load-if-needed path-str)
         (syntax (trms:in-process module-identifier
				  initial-required-module-name
				  (provide exported-id ...)
				  (require require-spec ...)
				  (import modname)
				  form ...)))))
     ;; require-spec that is an <identifier>
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       (require modname require-spec ...)
		       form ...)
      (identifier? (syntax modname))
      (syntax (trms:in-process module-identifier initial-required-module-name
			       (provide exported-id ...)
			       (require require-spec ...)
			       (import modname)
			       form ...)))
     ;; done if no require, deftype, or defrep at beginning
     ((trms:in-process module-identifier initial-required-module-name
		       (provide exported-id ...)
		       form ...)
      (syntax (module module-identifier (exported-id ...)
                      form ...)))
     )))

(define-syntax require
  (lambda (x)
    (syntax-case x ()
      ((require )
        (void))
      ;; the form that SLIB uses (require 'name)
      ((require (kwd name))
        (literal-identifier=? (syntax kwd) (syntax quote))
        (syntax (slib:require (quote name))))
      ;; the MzScheme forms follow
      ;; require-spec of the form (file path)
      ((require (kwd path) require-spec ...)
        (and (literal-identifier=? (syntax kwd) (syntax file))
	     (string? (syntax-object->datum (syntax path))))
        (let ((mod-sym
               (trms:extract-name
                (syntax-object->datum (syntax path))))
              (path-str (syntax-object->datum (syntax path))))
          (with-syntax ((modname (datum->syntax-object (syntax kwd) mod-sym)))
          ;; have to load at expansion time
           (trms:load-if-needed path-str)
           (syntax (begin (require require-spec ...) (import modname))))))
      ;; require-spec of the form (lib fname collect ...)
      ((require (kwd fname collect ...) require-spec ...)
        (and (literal-identifier=? (syntax kwd) (syntax lib))
	     (string? (syntax-object->datum (syntax fname)))
             (trms:all string?
                       (map syntax-object->datum (syntax (collect ...)))))
        (let ((mod-sym
               (trms:extract-name
                (syntax-object->datum (syntax fname))))
              (path-str
               (string-append
		*trms:collects-dir* "/"
                (trms:add-slashes
                 (syntax-object->datum (syntax (collect ...))))
                (syntax-object->datum (syntax fname)))))
          (with-syntax ((modname (datum->syntax-object (syntax kwd) mod-sym)))
          ;; have to load at expansion time
           (trms:load-if-needed path-str)
           (syntax (begin (require require-spec ...) (import modname))))))
      ;; require-spec of the form "path"
      ((require path require-spec ...)
        (string? (syntax-object->datum (syntax path)))
        (let ((mod-sym
               (trms:extract-name
                (syntax-object->datum (syntax path))))
              (path-str (syntax-object->datum (syntax path))))
          (with-syntax ((modname
                         (datum->syntax-object (syntax require) mod-sym)))
          ;; have to load at expansion time
           (trms:load-if-needed path-str)
           (syntax (begin (require require-spec ...) (import modname))))))
      ;; require-spec that is an <identifier>
      ((require modname require-spec ...)
        (identifier? (syntax modname))
        (syntax (begin (import modname)
                       (require require-spec ...))))
      )))

