;;; $Id: tc-hash-table.scm,v 1.8 2006/04/25 17:32:42 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Hash tables used for type environments and unification.
;;; These are built using the hash tables in slib.
;;;
;;; Note that it's important to have a hash that is *not* based on object
;;; addresses (for symbols), because *tc:global-var-types* will be read in
;;; from a file and if an object hash is used, then most global names won't
;;; be found.  So don't change "equal?" below to "eq?"

(module tc-hash-table (lib "typedscm.ss" "typedscm")

(provide tc:hash-table? tc:make-hash-table 
	tc:hash-table-defined? tc:hash-table-value
	tc:hash-table-associate tc:hash-table-associate*
	tc:hash-table-join
	tc:hash-table-map tc:hash-table-filter
	tc:hash-table-for-each)

(deftype tc:make-hash-table (forall (K V) (-> () (tc:hash-table K V))))
(deftype tc:hash-table-defined?
  (forall (K V) (-> (K (tc:hash-table K V)) boolean)))
(deftype tc:hash-table-value
  (forall (K V) (-> (K (tc:hash-table K V) (-> () V)) V)))
(deftype tc:hash-table-associate
  (forall (K V) (-> (K V (tc:hash-table K V)) (tc:hash-table K V))))
(deftype tc:hash-table-associate*
  (forall (K V)
     (-> ((list-of K) (list-of V) (tc:hash-table K V)) (tc:hash-table K V))))
(deftype tc:hash-table-join
  (forall (K V)
     (-> ((tc:hash-table K V) (tc:hash-table K V)) (tc:hash-table K V))))
(deftype tc:hash-table-map
  (forall (K S T)
     (-> ((-> (K S) T) (tc:hash-table K S))
	 (tc:hash-table K T))))
(deftype tc:hash-table-filter
  (forall (K V) (-> ((-> (K V) boolean) (tc:hash-table K V))
		    (tc:hash-table K V))))
(deftype tc:hash-table-for-each
  (forall (K V) (-> ((-> (K V) void) (tc:hash-table K V)) void)))

(require (lib "slib-hashtab.scm" "typedscm")
         (lib "relax-type-checking.scm" "typedscm"))

(define-datatype tc:hash-table tc:hash-table?
  (tc:hash-table-record (count number?) (ht slib:hash-table?)))

(define tc:hash-table->count
  (has-type-trusted
    (forall (K V) (-> ((tc:hash-table K V)) number))
    (lambda (htr)
      (cases tc:hash-table htr
	(tc:hash-table-record (count ht) count)))))

(define tc:hash-table->ht
  (has-type-trusted
    (forall (K V) (-> ((tc:hash-table K V)) (slib:hash-table K V)))
    (lambda (htr)
      (cases tc:hash-table (test-type tc:hash-table? htr)
	(tc:hash-table-record (count ht) ht)))))

(define tc:make-hash-table
  (lambda ()
    (has-type-trusted
      (forall (K V) (tc:hash-table K V))
      (tc:hash-table-record 0 (make-hash-table 1)))))

(define tc:hash-table-defined?
  (let ((equal-hash-inquirer? (hash-inquirer equal?)))
    (lambda (sym tcht)
      (equal-hash-inquirer? (tc:hash-table->ht tcht) sym))))

(define tc:hash-table-value
  (let ((equal-hash-value (hash-inquirer equal?)))
    (lambda (sym tcht failure-cont)
      (let ((val (equal-hash-value (tc:hash-table->ht tcht) sym)))
	(if (true? val)
	    val
	    (failure-cont))))))

(define tc:hash-table-associate
  (let ((equal-hash-associator! (hash-associator equal?))
	(equal-hash-asso (predicate->hash-asso equal?)))
    (lambda (sym val tcht)
      (let* ((count (tc:hash-table->count tcht))
	     (new-slht (tc:copy-hash-table (+ 1 count) tcht))
	     (res (equal-hash-asso sym new-slht)))
	(if (true? res)
	    (set-cdr! res val)
	    (ignore (equal-hash-associator! new-slht sym val)))
	(tc:hash-table-record
	 (+ count (if (true? res) 0 1))
	 new-slht)))))

(define tc:hash-table-associate*
  (let ((equal-hash-associator! (hash-associator equal?))
	(equal-hash-asso (predicate->hash-asso equal?)))
    (lambda (syms vals tcht)
      (let* ((new-count (tc:hash-table->count tcht))
	     (new-slht
	      (tc:copy-hash-table (+ new-count (length syms)) tcht)))
	(for-each (lambda (sym val)
		    (let ((res (equal-hash-asso sym new-slht)))
		      (if (true? res)
			  (set-cdr! res val)
			  (begin
			    (set! new-count (+ 1 new-count))
			    (ignore
			      (equal-hash-associator! new-slht sym val))))))
		  syms vals)
	(tc:hash-table-record new-count new-slht)))))

  (define tc:hash-table-join
    (let ((equal-hash-associator! (hash-associator equal?))
	  (equal-hash-inquirer? (hash-inquirer equal?)))
      (lambda (tcht1 tcht2)
	;; ENSURES: result is a new hash table containing the bindings
	;; in ht1 and ht2 (ht1 has precedence)
	(let* ((new-count (tc:hash-table->count tcht1))
	       (new-slht
		(tc:copy-hash-table (+ new-count (tc:hash-table->count tcht2))
				    tcht1))
	       (slht1 (tc:hash-table->ht tcht1)))
	  (hash-for-each
	   (lambda (sym val)
	     (if (not (true? (equal-hash-inquirer? slht1 sym)))
		 (begin
		   (set! new-count (+ 1 new-count))
		   (ignore (equal-hash-associator! new-slht sym val)))))
	   (tc:hash-table->ht tcht2))
	  (tc:hash-table-record new-count new-slht)))))
    
(define tc:hash-table-map
  (has-type-trusted
   (forall (K S T)
           (-> ((-> (K S) T) (tc:hash-table K S))
               (tc:hash-table K T)))
   (lambda (p tcht)
     (tc:hash-table-record
      (tc:hash-table->count tcht)
      (hash-map p (tc:hash-table->ht tcht))))))
      
(define tc:hash-table-filter
  (let ((equal-hash-associator! (hash-associator equal?)))
    (lambda (pred tcht)
      (let ((new-slht (make-hash-table
		       (tc:compute-new-size (tc:hash-table->count tcht))))
	    (new-count 0))
	(hash-for-each 
         (lambda (sym val)
           (if (pred sym val)
               (begin
                 (set! new-count (+ new-count 1))
                 (ignore (equal-hash-associator! new-slht sym val)))))
         (tc:hash-table->ht tcht))
	(tc:hash-table-record new-count new-slht)))))

(define tc:hash-table-for-each
  (lambda (p tcht)
     (hash-for-each p (tc:hash-table->ht tcht))))

(deftype tc:copy-hash-table
  (forall (K V) (-> (number (tc:hash-table K V)) (slib:hash-table K V))))
(define tc:copy-hash-table
  (let ((equal-hash-associator! (hash-associator equal?)))
    (lambda (new-count tcht)
      (let ((new-slht (make-hash-table (tc:compute-new-size new-count)))
	    (count (tc:hash-table->count tcht))
	    (old-slht (tc:hash-table->ht tcht)))
      	(hash-for-each (lambda (sym val)
			 (ignore
                          (equal-hash-associator! new-slht sym val)))
		       old-slht)
	new-slht))))
      
(deftype tc:compute-new-size (-> (number) number))
(define tc:compute-new-size
  (lambda (count)
    ;; REQUIRES: count is at least 0.
    ;; ENSURES: Result is a number of buckets appropriate for a hash table
    ;;          to store count items.
    (+ (* 2 count) 1)))
  
) ;; end module
