;;; $Id: tc-scheme-unparser.scm,v 1.27 2006/02/10 05:24:19 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Unparser for the Scheme abstract syntax, with additions for EOPL 2e,
;;; and our type checker.

(module tc-scheme-unparser (lib "typedscm.ss" "typedscm")

(provide tc:unparse-scheme-syntax tc:unparse-scheme-program
         tc:unparse-scheme-form
         tc:unparse-scheme-module-name tc:unparse-scheme-abs-conc-pair
         tc:unparse-scheme-definition tc:unparse-scheme-formals
         tc:unparse-scheme-body tc:unparse-variant-decl
         tc:unparse-field-pred-binding tc:unparse-scheme-syntax-spec
         tc:unparse-scheme-transformer-spec tc:unparse-scheme-syntax-rule
         tc:unparse-scheme-syntax-pattern tc:unparse-scheme-syntax-template
         tc:unparse-scheme-syntax-template-element tc:unparse-scheme-exp
         tc:unparse-scheme-cond-clause tc:unparse-scheme-case-clause
         tc:unparse-scheme-binding-spec tc:unparse-scheme-iteration-spec
         tc:unparse-scheme-cases-clause tc:unparse-scheme-datum
         tc:unparse-input-type-exp tc:unparse-variant-type-decl
         tc:unparse-field-type-binding
         )

;;; Main public procedures:
(deftype tc:unparse-scheme-program (-> (tc:scheme-program) datum))
(deftype tc:unparse-scheme-form (-> (tc:scheme-form) datum))
(deftype tc:unparse-scheme-exp (-> (tc:scheme-exp) datum))

(require (lib "tc-scheme-abstract-syntax.scm" "typedscm"))

;;; error handling
(deftype tc:unparse-error (-> (string datum) poof))
(define tc:unparse-error
  (lambda (name offending-datum)
    (error (string-append name ": bad abstract syntax argument:")
           offending-datum)))

(deftype tc:unparse-scheme-syntax
  (forall (syn) (-> (syn) datum)))
(define tc:unparse-scheme-syntax
  (lambda (e)
    (has-type-trusted ;; the type checker doesn't understand this pattern
      datum
      (let ((unparse
             (cond
              ((tc:scheme-form? e) tc:unparse-scheme-form)
              ((tc:scheme-exp? e) tc:unparse-scheme-exp)
              ((tc:scheme-datum? e) tc:unparse-scheme-datum)
              ((tc:scheme-formals? e) tc:unparse-scheme-formals)
              ((tc:scheme-body? e) tc:unparse-scheme-body)
              ((tc:scheme-definition? e) tc:unparse-scheme-definition)
              ((tc:variant-decl? e) tc:unparse-variant-decl)
              ((tc:field-pred-binding? e) tc:unparse-field-pred-binding)
              ((tc:scheme-cond-clause? e) tc:unparse-scheme-cond-clause)
              ((tc:scheme-case-clause? e) tc:unparse-scheme-case-clause)
              ((tc:scheme-binding-spec? e) tc:unparse-scheme-binding-spec)
              ((tc:scheme-cases-clause? e) tc:unparse-scheme-cases-clause)
              ((tc:scheme-iteration-spec? e) tc:unparse-scheme-iteration-spec)
              ((tc:scheme-program? e) tc:unparse-scheme-program)
              ((tc:scheme-module-name? e) tc:unparse-scheme-module-name)
              ((tc:scheme-abs-conc-pair? e) tc:unparse-scheme-abs-conc-pair)
              (else (tc:unparse-error "tc:unparse-scheme-syntax" e))
              )))      
        (unparse e)))))

;;; programs
(define tc:unparse-scheme-program
  (lambda (sp)
    (cases tc:scheme-program sp
      (tc:scheme-program-normal (position require-forms deftypes forms)
        (if (= 1 (length forms))
            (tc:unparse-scheme-form (car forms))
            (cons 'begin
                  (append (map tc:unparse-scheme-form require-forms)
                          (map tc:unparse-scheme-definition deftypes)
                          (map tc:unparse-scheme-form forms)))))
      )))

;;; We use has-type-trusted in this file from this point below mainly
;;; because type checking doesn't work and is too slow on code with
;;; lots of cons, car, cdr stuff.

;;; forms (<command or definition>s)
(define tc:unparse-scheme-form
  (lambda (form)
    (has-type-trusted
      datum
      (cases tc:scheme-form form
        (tc:command (position exp) (tc:unparse-scheme-exp exp))
        (tc:load-command (position operator filename program)
         (list operator filename))
        (tc:definition (position definition)
         (tc:unparse-scheme-definition definition))
        (tc:syntax-definition (position spec)
         (cons 'define-syntax (tc:unparse-scheme-syntax-spec spec)))
        (tc:begin-form (position forms)
         (cons 'begin (map tc:unparse-scheme-form forms)))
        (tc:module (position identifier initial-required-module-name
                           provides requires-for-syntax requires defreps
                           deftypes definitions)
         (append
          (list 'module identifier)
          (list (tc:unparse-scheme-module-name initial-required-module-name))
          (list (cons 'provide provides))
          (map tc:unparse-scheme-definition deftypes)
          (if (null? requires-for-syntax)
              '()
              (list (cons 'require-for-syntax
                          (map tc:unparse-scheme-module-name requires-for-syntax))))      
          (if (null? requires)
              '()
              (list (cons 'require
                          (map tc:unparse-scheme-module-name requires))))
          (if (null? defreps)
              '()
              (list (cons 'defrep
                          (map tc:unparse-scheme-abs-conc-pair defreps))))
          (map tc:unparse-scheme-definition definitions)))
        (tc:require-form (position module-names)
          (cons 'require (map tc:unparse-scheme-module-name module-names)))
	(else
	 (tc:unparse-error "tc:unparse-scheme-form" form))))))

;;; module-names
(deftype tc:unparse-scheme-module-name
  (-> (tc:scheme-module-name) datum))
(define tc:unparse-scheme-module-name
  (lambda (mn)
    (has-type-trusted
      datum
      (cases tc:scheme-module-name mn
	(tc:module-id (position identifier) identifier)
	(tc:unix-relative-path-string (position path) path)
	(tc:file (position path) (list 'file path))
	(tc:lib
	 (position filename collects)
	 (append (list 'lib filename) collects))
	(tc:import-from-module-name
	 (name)
	 (tc:unparse-scheme-module-name name))
	(else
	 (tc:unparse-error "tc:unparse-scheme-module-name" mn))))))

;;; abs-conc-pairs
(deftype tc:unparse-scheme-abs-conc-pair
  (-> (tc:scheme-abs-conc-pair) datum))
(define tc:unparse-scheme-abs-conc-pair
  (lambda (acp)
    (has-type-trusted
      datum
      (cases tc:scheme-abs-conc-pair acp
	(tc:abs-conc-pair-simple (position abstract-type concrete-type)
	 (list (tc:unparse-input-type-exp abstract-type)
	       (tc:unparse-input-type-exp concrete-type)))
	(tc:abs-conc-pair-polymorphic
	 (position type-vars abstract-type concrete-type)
         (list 'forall type-vars
               (tc:unparse-input-type-exp abstract-type)
               (tc:unparse-input-type-exp concrete-type)))
	(else
	 (tc:unparse-error "tc:unparse-scheme-abs-conc-pair" acp))))))

;;; definitions
(deftype tc:unparse-scheme-definition
  (-> (tc:scheme-definition) datum))
(define tc:unparse-scheme-definition
  (lambda (definition)
    (has-type-trusted
      datum
      (cases tc:scheme-definition definition
	(tc:simple-define (position variable exp)
	  (list 'define variable (tc:unparse-scheme-exp exp)))
	(tc:procedure-define (position variable formals body)
	  (append
	   (list 'define
		 (cons variable (tc:unparse-scheme-formals formals)))
	   (tc:unparse-scheme-body body)))
	(tc:grouped-define (position definitions)
	 (cons 'begin (map tc:unparse-scheme-definition definitions)))
	(tc:define-datatype
	 (position type-name type-predicate-name variants)
         (cons 'define-datatype
               (cons type-name
                     (cons type-predicate-name
                           (map tc:unparse-variant-decl variants)))))
	(tc:deftype (position variable type)
	 (list 'deftype variable (tc:unparse-input-type-exp type)))
	(else
	 (tc:unparse-error "tc:unparse-scheme-definition" definition))))))

;;; formals
(deftype tc:unparse-scheme-formals
  (-> (tc:scheme-formals) datum))
(define tc:unparse-scheme-formals
  (lambda (sformals)
    (has-type-trusted
      datum 
      (cases tc:scheme-formals sformals
	(tc:normal-formals (position variables) variables)
	(tc:unrestricted-formals (position variable) variable)
	(tc:some-normal-formals (position before-dot after-dot)
	 ;; append actually works with the last argument not a list,
	 ;; and makes an improper list as desired in this case
	 (append before-dot after-dot))
      (else
       (tc:unparse-error "tc:unparse-scheme-formals" sformals))))))

;;; bodies
(deftype tc:unparse-scheme-body
  (-> (tc:scheme-body-record) (list-of datum)))
(define tc:unparse-scheme-body
  (lambda (body)
    (cases tc:scheme-body-record body
      (tc:scheme-body (position definitions sequence)
         (append (map tc:unparse-scheme-definition definitions)
                 (list (tc:unparse-scheme-exp sequence))))
      (else
       (tc:unparse-error "tc:unparse-scheme-body" body)))))

;;; variant-decls
(deftype tc:unparse-variant-decl
  (-> (tc:variant-decl-record) datum))
(define tc:unparse-variant-decl
  (lambda (variant)
    (cases tc:variant-decl-record variant
      (tc:variant-decl (position variant-name fields)
	(cons variant-name (map tc:unparse-field-pred-binding fields)))
      (else
       (tc:unparse-error "tc:unparse-variant-decl" variant)))))

;;; field-pred binding
(deftype tc:unparse-field-pred-binding
  (-> (tc:field-pred-binding-record) datum))
(define tc:unparse-field-pred-binding
  (lambda (binding)
    (has-type-trusted
      datum
      (cases tc:field-pred-binding-record binding
        (tc:field-pred-binding (position field-name predicate-exp)
           (list field-name (tc:unparse-scheme-exp predicate-exp)))
        (else
           (tc:unparse-error "tc:unparse-field-pred-binding" binding))))))

;;; syntax specs
(deftype tc:unparse-scheme-syntax-spec
  (-> (tc:scheme-syntax-spec-record) datum))
(define tc:unparse-scheme-syntax-spec
  (lambda (sss)
    (cases tc:scheme-syntax-spec-record sss
      (tc:scheme-syntax-spec (position keyword transformer)
         (list keyword (tc:unparse-scheme-transformer-spec transformer)))
      (else
         (tc:unparse-error "tc:unparse-scheme-syntax-spec" sss)))))

;;; transformer specs
(deftype tc:unparse-scheme-transformer-spec
  (-> (tc:scheme-transformer-spec-record) datum))
(define tc:unparse-scheme-transformer-spec
  (lambda (transformer)
    (cases tc:scheme-transformer-spec-record transformer
      (tc:scheme-transformer-spec (position identifiers syntax-rules)
         (cons
          'syntax-rules
          (cons
           identifiers
           (map tc:unparse-scheme-syntax-rule syntax-rules))))
     (else
        (tc:unparse-error "tc:unparse-scheme-transformer-spec" transformer)))))

;;; syntax rules
(deftype tc:unparse-scheme-syntax-rule
  (-> (tc:scheme-syntax-rule-record) datum))
(define tc:unparse-scheme-syntax-rule
  (lambda (sr)
    (cases tc:scheme-syntax-rule-record sr
      (tc:scheme-syntax-rule (position pattern template)
         (list (tc:unparse-scheme-syntax-pattern pattern)
               (tc:unparse-scheme-syntax-template template)))
     (else
      (tc:unparse-error "tc:unparse-scheme-syntax-rule" sr)))))

;;; syntax patterns
(deftype tc:unparse-scheme-syntax-pattern
  (-> (tc:scheme-syntax-pattern) datum))
(define tc:unparse-scheme-syntax-pattern
  (lambda (pat)
    (has-type-trusted
      datum
      (cases tc:scheme-syntax-pattern pat
	(tc:syntax-pattern-identifier (position identifier) identifier)
	(tc:syntax-pattern-normal-list
	 (position patterns)
	 (map tc:unparse-scheme-syntax-pattern patterns))
	(tc:syntax-pattern-dotted-list
	 (position before-dot after-dot)
	 (append (map tc:unparse-scheme-syntax-pattern before-dot)
		 (tc:unparse-scheme-syntax-pattern after-dot)))
	(tc:syntax-pattern-ellipsis-list
	 (position patterns repeated)
	 (append (map tc:unparse-scheme-syntax-pattern patterns)
                 (cons (tc:unparse-scheme-syntax-pattern repeated)
                       (list '...))))
	(tc:syntax-pattern-normal-vector
	 (position patterns)
	 (list->vector (map tc:unparse-scheme-syntax-pattern patterns)))
	(tc:syntax-pattern-ellipsis-vector
	 (position patterns repeated)
	 (list->vector (append
			(map tc:unparse-scheme-syntax-pattern patterns)
			(list '...))))
	(tc:syntax-pattern-datum (position datum) datum)
	(else
	 (tc:unparse-error "tc:unparse-scheme-syntax-pattern" pat))))))

;;; syntax templates
(deftype tc:unparse-scheme-syntax-template
  (-> (tc:scheme-syntax-template) datum))
(define tc:unparse-scheme-syntax-template
  (lambda (template)
    (has-type-trusted
      datum
      (cases tc:scheme-syntax-template template
	(tc:syntax-template-identifier (position identifier) identifier)
	(tc:syntax-template-normal-list
	 (position elements)
	 (apply append
                (map tc:unparse-scheme-syntax-template-element elements)))
	(tc:syntax-template-dotted-list
	 (position before-dot after-dot)
	 (append
	  (apply append
		 (map tc:unparse-scheme-syntax-template-element before-dot))
	  (tc:unparse-scheme-syntax-template-element after-dot)))
	(tc:syntax-template-normal-vector
	 (position elements)
	 (list->vector
	  (apply append
		 (map tc:unparse-scheme-syntax-template-element elements))))
	(tc:syntax-template-datum (position datum) datum)
	(else
	 (tc:unparse-error "tc:unparse-scheme-syntax-template" template))))))


;;; syntax template elements
(deftype tc:unparse-scheme-syntax-template-element
  (-> (tc:scheme-syntax-template-element) datum))
(define tc:unparse-scheme-syntax-template-element
  (lambda (te)
    (has-type-trusted
      datum
      (cases tc:scheme-syntax-template-element te
	(tc:scheme-syntax-template-element-normal
	 (position element)
	 (list (tc:unparse-scheme-syntax-template element)))
	(tc:scheme-syntax-template-element-ellipsis
	 (position element)
	 (list (tc:unparse-scheme-syntax-template element)
	       '...))
	(else
	 (tc:unparse-error "tc:unparse-scheme-syntax-template-element" te))))))

;;; expressions
(define tc:unparse-scheme-exp
  (lambda (expr)
    (has-type-trusted
      datum
      (cases tc:scheme-exp expr
        (tc:varref (position variable) variable)
        (tc:quotation (position datum)
          (list 'quote (tc:unparse-scheme-datum datum)))
        (tc:self-evaluating (position datum) datum)
        (tc:procedure-call (position operator operands)
          (cons (tc:unparse-scheme-exp operator)
                (map tc:unparse-scheme-exp operands)))
        (tc:lambda-exp (position formals body)
          (append
           (list 'lambda (tc:unparse-scheme-formals formals))
           (tc:unparse-scheme-body body)))
        (tc:named-lambda-exp (position variable formals body)
          (append
           (list 'lambda (tc:unparse-scheme-formals formals))
           (tc:unparse-scheme-body body)))
        (tc:conditional-exp (position test consequent alternate)
          (list 'if
                (tc:unparse-scheme-exp test)
                (tc:unparse-scheme-exp consequent)
                (tc:unparse-scheme-exp alternate)))
        (tc:conditional-stmt (position test consequent)
          (list 'if (tc:unparse-scheme-exp test)
                (tc:unparse-scheme-exp consequent)))
        (tc:assignment (position variable exp)
          (list 'set! variable (tc:unparse-scheme-exp exp)))
        (tc:macro-use (position keyword datums)
          (cons keyword
                (map tc:unparse-scheme-datum datums)))
        (tc:let-syntax-exp (position syntax-specs body)
          (append
           (list 'let-syntax
                 (map tc:unparse-scheme-syntax-spec syntax-specs))
           (tc:unparse-scheme-body body)))
        (tc:letrec-syntax-exp (position syntax-specs body)
          (append
           (list 'letrec-syntax
                 (map tc:unparse-scheme-syntax-spec syntax-specs))
           (tc:unparse-scheme-body body)))
        ;; derived expressions
        (tc:cond-exp (position clauses)
          (cons 'cond
                (map tc:unparse-scheme-cond-clause clauses)))
        (tc:cond-stmt (position clauses)
          (cons 'cond
                (map tc:unparse-scheme-cond-clause clauses)))
        (tc:case-exp (position exp clauses)
          (cons
           'case
           (cons (tc:unparse-scheme-exp exp)
                 (map tc:unparse-scheme-case-clause clauses))))
        (tc:case-stmt (position exp clauses)
          (cons
           'case
           (cons (tc:unparse-scheme-exp exp)
                 (map tc:unparse-scheme-case-clause clauses))))
        (tc:and-exp (position tests)
          (cons 'and (map tc:unparse-scheme-exp tests)))
        (tc:or-exp (position tests)
          (cons 'or (map tc:unparse-scheme-exp tests)))
        (tc:let-exp (position decls body)
          (append
           (list 'let
                 (map tc:unparse-scheme-binding-spec decls))
           (tc:unparse-scheme-body body)))
        (tc:named-let-exp (position variable decls body)
          (append
           (list 'let
                 variable
                 (map tc:unparse-scheme-binding-spec decls))
           (tc:unparse-scheme-body body)))
        (tc:let*-exp (position decls body)
          (append
           (list 'let*
                 (map tc:unparse-scheme-binding-spec decls))
           (tc:unparse-scheme-body body)))
        (tc:letrec-exp (position decls body)
          (append
           (list 'letrec
                 (map tc:unparse-scheme-binding-spec decls))
           (tc:unparse-scheme-body body)))
        (tc:begin-exp (position commands exp)
          (cons 'begin
                (append (map tc:unparse-scheme-exp commands)
                        (list (tc:unparse-scheme-exp exp)))))
        (tc:do-exp (position iteration-specs test do-result commands)
          (append
           (list 'do
                 (map tc:unparse-scheme-iteration-spec iteration-specs)
                 (list (tc:unparse-scheme-exp test)
                       (tc:unparse-scheme-exp do-result)))
           (map tc:unparse-scheme-exp commands)))
        (tc:do-stmt (position iteration-specs test commands)
          (append
           (list 'do
                 (map tc:unparse-scheme-iteration-spec iteration-specs)
                 (list (tc:unparse-scheme-exp test)))
           (map tc:unparse-scheme-exp commands)))
        (tc:delay-exp (position exp)
          (list 'delay (tc:unparse-scheme-exp exp)))
        (tc:quasiquotation (position datum)
          (list 'quasiquote (tc:unparse-scheme-datum datum)))
        ;; the following expressions are added for EOPL (second edition)
        (tc:cases-exp (position type-name exp clauses)
          (append
           (list 'cases
                 type-name
                 (tc:unparse-scheme-exp exp))
           (map tc:unparse-scheme-cases-clause clauses)))
        (tc:cases-stmt (position type-name exp clauses)
          (append
           (list 'cases
                 type-name
                 (tc:unparse-scheme-exp exp))
           (map tc:unparse-scheme-cases-clause clauses)))
        ;; the following expressions are added for the type checker
        (tc:has-type-exp (position type exp)
          (list 'has-type
                (tc:unparse-input-type-exp type)
                (tc:unparse-scheme-exp exp)))
        (tc:has-type-trusted-exp (position type exp)
          (list 'has-type-trusted
                (tc:unparse-input-type-exp type)
                (tc:unparse-scheme-exp exp)))
        (tc:test-type-exp (position type-pred exp)
          (list 'test-type
                (tc:unparse-scheme-exp type-pred)
                (tc:unparse-scheme-exp exp)))
        (tc:arrow-exp (position arg-exps result-exp)
          (list '->
                (map tc:unparse-scheme-exp arg-exps)
                (tc:unparse-scheme-exp result-exp)))
        (tc:varargs-arrow-exp (position arg-exps repeated-exp result-exp)
          (list '->
                (append (map tc:unparse-scheme-exp arg-exps)
                        (cons (tc:unparse-scheme-exp repeated-exp)
                              (cons '... '())))
                (tc:unparse-scheme-exp result-exp)))
        (tc:forall-exp (position formals exp)
          (list 'forall
                (tc:unparse-scheme-formals formals)
                (tc:unparse-scheme-exp exp)))
        (tc:define-datatype-exp
         (position variants)
         (cons 'define-datatype (map tc:unparse-variant-decl variants)))
        (tc:constructor-for-variant-type-exp
         (position argument-preds type-name)
         (cons 'constructor-for-variant
               (cons (map tc:unparse-scheme-exp argument-preds)
                     (list type-name))))
        (tc:module-exp (position identifier initial-required-module-name
                                 provides requires-for-syntax requires defreps
                                 deftypes definitions)
           (append
            (list 'module identifier)
            (list (tc:unparse-scheme-module-name initial-required-module-name))
            (list (cons 'provide provides))
            (map tc:unparse-scheme-definition deftypes)
            (if (null? requires-for-syntax)
                '()
                (list (cons 'require-for-syntax
                            (map tc:unparse-scheme-module-name requires-for-syntax))))
            (if (null? requires)
                '()
                (list (cons 'require
                            (map tc:unparse-scheme-module-name requires))))
            (if (null? defreps)
                '()
                (list (cons 'defrep
                            (map tc:unparse-scheme-abs-conc-pair defreps))))
            (map tc:unparse-scheme-definition definitions)))
  	(else
  	 (tc:unparse-error "tc:unparse-scheme-exp" expr))))))


;;; cond clause
(deftype tc:unparse-scheme-cond-clause
  (-> (tc:scheme-cond-clause) datum))
(define tc:unparse-scheme-cond-clause
  (lambda (clause)
    (has-type-trusted
      datum
      (cases tc:scheme-cond-clause clause
        (tc:normal-cond-clause (position test sequence)
           (list (tc:unparse-scheme-exp test)
  	       (tc:unparse-scheme-exp sequence)))
        (tc:test-only-cond-clause (position test)
           (list (tc:unparse-scheme-exp test)))
        (tc:test=>cond-clause (position test recipient)
           (list (tc:unparse-scheme-exp test)
                 '=>
                 (tc:unparse-scheme-exp recipient)))
        (tc:else-cond-clause (position sequence)
           (list 'else (tc:unparse-scheme-exp sequence)))
        (else
    	 (tc:unparse-error "tc:unparse-scheme-cond-clause" clause))))))

;;; case clause
(deftype tc:unparse-scheme-case-clause
  (-> (tc:scheme-case-clause) datum))
(define tc:unparse-scheme-case-clause
  (lambda (clause)
    (has-type-trusted
      datum
      (cases tc:scheme-case-clause clause
        (tc:normal-case-clause (position datums sequence)
           (list (map tc:unparse-scheme-datum datums)
		 (tc:unparse-scheme-exp sequence)))
        (tc:else-case-clause (position sequence)
           (list 'else (tc:unparse-scheme-exp sequence)))
        (else
         (tc:unparse-error "tc:unparse-scheme-case-clause" clause))))))

;;; binding spec (for let, let*, letrec)
(deftype tc:unparse-scheme-binding-spec
  (-> (tc:scheme-binding-spec-record) datum))
(define tc:unparse-scheme-binding-spec
  (lambda (binding)
    (cases tc:scheme-binding-spec-record binding
      (tc:scheme-binding-spec (position variable exp)
         (list variable (tc:unparse-scheme-exp exp)))
      (else
       (tc:unparse-error "tc:unparse-scheme-binding-spec" binding)))))

;;; iteration spec (for do)
(deftype tc:unparse-scheme-iteration-spec
  (-> (tc:scheme-iteration-spec) datum))
(define tc:unparse-scheme-iteration-spec
  (lambda (is)
    (has-type-trusted
      datum
      (cases tc:scheme-iteration-spec is
        (tc:stepped-iteration-spec (position variable init step)
	   (list variable
		 (tc:unparse-scheme-exp init)
		 (tc:unparse-scheme-exp step)))
        (tc:init-only-iteration-spec (position variable init)
           (list variable
                 (tc:unparse-scheme-exp init)))
        (else
         (tc:unparse-error "tc:unparse-scheme-binding-spec" is))))))

;;; cases clause
(deftype tc:unparse-scheme-cases-clause
  (-> (tc:scheme-cases-clause) datum))
(define tc:unparse-scheme-cases-clause
  (lambda (clause)
    (has-type-trusted
      datum
      (cases tc:scheme-cases-clause clause
        (tc:normal-cases-clause (position record-name fields sequence)
	   (list record-name fields (tc:unparse-scheme-exp sequence)))
        (tc:else-cases-clause (position sequence)
           (list 'else (tc:unparse-scheme-exp sequence)))
	(else
	   (tc:unparse-error "tc:unparse-scheme-cases-clause" clause))))))

;;; scheme-datum (for quotations and quasiquotations)
(deftype tc:unparse-scheme-datum
  (-> (tc:scheme-datum) datum))
(define tc:unparse-scheme-datum
  (lambda (dat)
    (has-type-trusted
      datum
      (cases tc:scheme-datum dat
        (tc:simple-datum (position datum) datum)
        (tc:normal-list-datum (position elements)
           (map tc:unparse-scheme-datum elements))
        (tc:improper-list-datum (position before-dot after-dot)
           (append (map tc:unparse-scheme-datum before-dot)
                   (tc:unparse-scheme-datum after-dot)))
        (tc:quoted-datum (position datum)
           (list 'quote (tc:unparse-scheme-datum datum)))
        (tc:quasiquoted-datum (position datum)
           (list 'quasiquote (tc:unparse-scheme-datum datum)))
        (tc:unquoted-datum (position datum)
           (list 'unquote (tc:unparse-scheme-datum datum)))
        (tc:unquoted-exp (position exp)
           (list 'unquote (tc:unparse-scheme-exp exp)))
        (tc:unquote-splicing-datum (position datum)
           (list 'unquote-splicing (tc:unparse-scheme-datum datum)))
        (tc:unquote-splicing-exp (position exp)
           (list 'unquote-splicing (tc:unparse-scheme-exp exp)))
        (tc:vector-datum (position elements)
           (apply vector (map tc:unparse-scheme-datum elements)))
        (else
         (tc:unparse-error "tc:unparse-scheme-datum" dat))))))

;;; input type expressions added for the type checker
(deftype tc:unparse-input-type-exp
  (-> (tc:input-type-exp) datum))
(define tc:unparse-input-type-exp
  (lambda (ityp)
    (has-type-trusted
      datum
      (cases tc:input-type-exp ityp
        (tc:basic-input-type-exp (position identifier) identifier)
        (tc:function-input-type-exp (position arg-types result-type)
           (list '->
                 (map tc:unparse-input-type-exp arg-types)
                 (tc:unparse-input-type-exp result-type)))
        (tc:varargs-function-input-type-exp
         (position arg-types repeated-type result-type)
           (list '->
                 (append (map tc:unparse-input-type-exp arg-types)
                         (list (tc:unparse-input-type-exp repeated-type)
                               '...))
                 (tc:unparse-input-type-exp result-type)))
        (tc:applied-input-type-exp (position operator operand-types)
           (cons operator
                 (map tc:unparse-input-type-exp operand-types)))
        (tc:polymorphic-input-type-exp (position type-formals monomorphic-type)
          (list 'forall
                type-formals
                (tc:unparse-input-type-exp monomorphic-type)))
        (tc:intersection-input-type-exp (position conjoined-types)
           (cons 'all-of (map tc:unparse-input-type-exp conjoined-types)))
        (tc:type-predicate-for-input-type-exp (position type)
           (list 'type-predicate-for (tc:unparse-input-type-exp type)))
        (tc:variant-record-input-type-exp (position variants)
           (cons 'variant-record (map tc:unparse-variant-type-decl variants)))
        (tc:translated-input-type-exp (position type-expr)
           (error "tc:unparse-input-type-exp: "
  		"Can't unparse this translated-input-type-exp " type-expr))
        (else
           (tc:unparse-error "tc:unparse-input-type-exp" ityp))))))

;;; variant-type-decls
(deftype tc:unparse-variant-type-decl
  (-> (tc:variant-type-decl-record) datum))
(define tc:unparse-variant-type-decl
  (lambda (variant)
    (cases tc:variant-type-decl-record variant
      (tc:variant-type-decl (position variant-name fields)
         (cons variant-name (map tc:unparse-field-type-binding fields)))
      (else
	 (tc:unparse-error "tc:unparse-variant-type-decl" variant)))))

;;; field-type binding
(deftype tc:unparse-field-type-binding
  (-> (tc:field-type-binding-record) datum))
(define tc:unparse-field-type-binding
  (lambda (binding)
    (cases tc:field-type-binding-record binding
      (tc:field-type-binding (position field-name type)
         (list field-name (tc:unparse-input-type-exp type)))
      (else
	 (tc:unparse-error "tc:unparse-variant-type-decl" binding)))))

) ;; end module
